
"""
Provides the #ObjectTypeModule that implements the de/serialization of JSON payloads for databind
schemas (see #databind.core.schema).
"""

import typing as t
from databind.core import annotations as A, Context, ConversionError, Direction, Converter, MapType, ObjectType


class ObjectTypeConverter(Converter):

  def _serialize(self, ctx: Context, type_: ObjectType) -> t.Dict[str, t.Any]:
    skip_default_values = True

    if not isinstance(ctx.value, type_.schema.python_type):
      raise ctx.type_error(expected=type_.schema.python_type)

    groups: t.Dict[str, t.Dict[str, t.Any]] = {}
    for field in type_.schema.fields.values():
      if not field.flat:
        continue
      value = getattr(ctx.value, field.name)
      if skip_default_values and value == field.get_default():
        continue
      groups[field.name] = ctx.push(field.type, value, field.name, field).convert()

    result: t.Dict[str, t.Any] = {}
    flattened = type_.schema.flattened()
    for name, flat_field in flattened.fields.items():
      alias = (flat_field.field.aliases or [name])[0]
      if not flat_field.group:
        value = getattr(ctx.value, name)
        if skip_default_values and value != flat_field.field.get_default():
          result[alias] = ctx.push(flat_field.field.type, value, name, flat_field.field).convert()
      elif alias in groups[flat_field.group or '$']:
        # May not be contained if we skipped default values.
        result[alias] = groups[flat_field.group or '$'][alias]

    # Explode values from the remainder field into the result.
    if flattened.remainder_field:
      assert isinstance(flattened.remainder_field.type, MapType)
      remnants = ctx.push(flattened.remainder_field.type, getattr(ctx.value, flattened.remainder_field.name), None, flattened.remainder_field).convert()
      for key, value in remnants.items():
        if key in result:
          raise ctx.error(f'key {key!r} of remainder field {flattened.remainder_field.name!r} cannot be exploded '
            'into resulting JSON object because of a conflict.')
        result[key] = value

    return result

  def _deserialize(self, ctx: Context, type_: ObjectType) -> t.Any:
    enable_unknowns = ctx.settings.get(A.enable_unknowns)
    typeinfo = ctx.get_annotation(A.typeinfo)

    if not isinstance(ctx.value, t.Mapping):
      raise ctx.type_error(expected=t.Mapping)

    groups: t.Dict[str, t.Dict[str, t.Any]] = {'$': {}}

    # Collect keys into groups.
    used_keys: t.Set[str] = set()
    flattened = type_.schema.flattened()
    for name, flat_field in flattened.fields.items():
      aliases = flat_field.field.aliases or [name]
      for alias in aliases:
        if alias in ctx.value:
          value = ctx.value[alias]
          groups.setdefault(flat_field.group or '$', {})[name] = \
            ctx.push(flat_field.field.type, value, name, flat_field.field).convert()
          used_keys.add(alias)
          break

    # Move captured groups into the root group ($).
    for group, values in groups.items():
      if group == '$': continue
      field = type_.schema.fields[group]
      groups['$'][group] = ctx.push(field.type, values, group, field).convert()

    # Collect unknown fields into the remainder field if there is one.
    if flattened.remainder_field:
      assert isinstance(flattened.remainder_field.type, MapType)
      remanants = {k: ctx.value[k] for k in ctx.value.keys() - used_keys}
      groups['$'][flattened.remainder_field.name] = ctx.push(flattened.remainder_field.type, remanants, None, flattened.remainder_field).convert()
      used_keys.update(ctx.value.keys())

    if not enable_unknowns or (enable_unknowns and enable_unknowns.callback):
      unused_keys = ctx.value.keys() - used_keys
      if unused_keys and not enable_unknowns:
        raise ConversionError(f'unknown keys found while deserializing {ctx.type}: {unused_keys}', ctx.location)
      elif unused_keys and enable_unknowns and enable_unknowns.callback:
        enable_unknowns.callback(ctx, set(unused_keys))

    try:
      return ((typeinfo.deserialize_as if typeinfo else None) or type_.schema.python_type)(**groups['$'])
    except TypeError as exc:
      raise ctx.error(str(exc))

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, ObjectType)
    if ctx.direction == Direction.serialize:
      return self._serialize(ctx, ctx.type)
    elif ctx.direction == Direction.deserialize:
      return self._deserialize(ctx, ctx.type)
    assert False




"""
Provides the #PlainDatatypeModule which contains converters for plain datatypes.
"""

import base64
import typing as t
from databind.core import annotations as A, ConcreteType, Context, Direction, Converter
from nr import preconditions


def _int_lossless(v: float) -> int:
  """ Convert *v* to an integer only if the conversion is lossless, otherwise raise an error. """
  preconditions.check_argument(v % 1.0 == 0.0, lambda: f'expected int, got {v!r}')
  return int(v)


def _bool_from_str(s: str) -> bool:
  """ Converts *s* to a boolean value based on common truthy keywords. """
  if s.lower() in ('yes', 'true', 'on', 'enabled'):
    return True
  if s.lower() in ('no', 'false', 'off', 'disabled'):
    return True
  raise ValueError(f'not a truthy keyword: {s!r}')


class PlainJsonConverter(Converter):
  """
  Converter for the following plain types:

  * #bool
  * #int
  * #float
  * #str

  Non-strict handling will allow conversion of values from str to other plain types during
  deserialization but not during serialization.
  """

  # Map for (source_type, target_type)
  _strict_adapters: t.Dict[t.Tuple[t.Type, t.Type], t.Callable[[t.Any], t.Any]] = {
    (bytes, bytes): lambda d: base64.b64encode(d).decode('ascii'),
    (str, bytes): base64.b64decode,
    (str, str): str,
    (int, int): int,
    (float, float): float,
    (int, float): float,
    (float, int): _int_lossless,
    (bool, bool): bool,
  }

  # Used only during deserialization if the #fieldinfo.strict is disabled.
  _nonstrict_adapters = _strict_adapters.copy()
  _nonstrict_adapters.update({
    (str, int): int,
    (str, float): float,
    (str, bool): _bool_from_str,
    (int, str): str,
    (float, str): str,
    (bool, str): str,
  })

  def convert(self, ctx: Context) -> t.Any:
    source_type = type(ctx.value)
    target_type = preconditions.check_instance_of(ctx.location.type, ConcreteType).type
    fieldinfo = ctx.get_annotation(A.fieldinfo) or A.fieldinfo()
    strict = ctx.direction == Direction.serialize or fieldinfo.strict
    func = (self._strict_adapters if strict else self._nonstrict_adapters)\
        .get((source_type, target_type))
    if func is None:
      raise ctx.error(f'unable to {ctx.direction.name} {source_type.__name__} -> {target_type.__name__}')
    try:
      return func(ctx.value)
    except ValueError as exc:
      raise ctx.error(str(exc))




"""
Implements de/serialization of the #datetime types #datetime.date, #datetime.datetime and
#datetime.time to/from strings. The date/time format can be specified using the #A.datefmt()
annotation.

The date is parsed using the #nr.parsing.date module.
"""

import datetime
import typing as t
from databind.core import annotations as A
from databind.core import  Context, Converter, ConcreteType, Direction
from nr import preconditions
from nr.parsing.date import ISO_8601, duration

T_DateTypes = t.TypeVar('T_DateTypes', bound=t.Union[datetime.date, datetime.time, datetime.datetime])


class DatetimeJsonConverter(Converter):

  DEFAULT_DATE_FMT = A.datefmt(ISO_8601)
  DEFAULT_TIME_FMT = A.datefmt(ISO_8601)
  DEFAULT_DATETIME_FMT = A.datefmt(ISO_8601)

  def convert(self, ctx: Context) -> t.Any:
    preconditions.check_instance_of(ctx.type, ConcreteType)
    type_ = t.cast(ConcreteType, ctx.type).type
    datefmt = ctx.get_annotation(A.datefmt) or (
      self.DEFAULT_DATE_FMT if type_ == datetime.date else
      self.DEFAULT_TIME_FMT if type_ == datetime.time else
      self.DEFAULT_DATETIME_FMT if type_ == datetime.datetime else None)
    assert datefmt is not None

    if ctx.direction == Direction.deserialize:
      if not isinstance(ctx.value, str):
        raise ctx.type_error(expected='str')
      dt = datefmt.parse(type_, ctx.value)  # TODO(NiklasRosenstein): Rethrow as ConversionError
      assert isinstance(dt, type_)
      return dt

    else:
      if not isinstance(ctx.value, type_):
        raise ctx.type_error(expected=type_)
      return datefmt.format(ctx.value)


class DurationConverter(Converter):

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, ConcreteType)

    if ctx.direction == Direction.serialize:
      if not isinstance(ctx.value, duration):
        raise ctx.type_error(expected=duration)
      return str(ctx.value)

    elif ctx.direction == Direction.deserialize:
      if not isinstance(ctx.value, str):
        raise ctx.type_error(expected=str)
      return duration.parse(ctx.value)   # TODO (@NiklasRosenstein): Reraise as ConversionError?




import typing as t
from databind.core import annotations as A, BaseType, Context, ConversionError, Converter, Direction, UnionType


class UnionConverter(Converter):

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, UnionType)
    fallback = ctx.get_annotation(A.union) or A.union()
    style = ctx.type.style or fallback.style or UnionType.DEFAULT_STYLE
    discriminator_key = ctx.type.discriminator_key or fallback.discriminator_key or UnionType.DEFAULT_DISCRIMINATOR_KEY

    is_deserialize = ctx.direction == Direction.deserialize

    if is_deserialize:
      if not isinstance(ctx.value, t.Mapping):
        raise ctx.type_error(expected=t.Mapping)
      if discriminator_key not in ctx.value:
        raise ConversionError(f'missing discriminator key {discriminator_key!r}', ctx.location)
      member_name = ctx.value[discriminator_key]
      member_type = ctx.type.subtypes.get_type_by_name(member_name, ctx.type_hint_adapter)
      assert isinstance(member_type, BaseType), f'"{type(ctx.type.subtypes).__name__}" returned member_type must '\
          f'be BaseType, got "{type(member_type).__name__}"'
    else:
      member_type = ctx.type_hint_adapter.adapt_type_hint(type(ctx.value))
      member_name = ctx.type.subtypes.get_type_name(member_type, ctx.type_hint_adapter)

    type_hint = member_type

    if is_deserialize:
      if style == A.union.Style.nested:
        if member_name not in ctx.value:
          raise ConversionError(f'missing union value key {member_name!r}', ctx.location)
        child_context = ctx.push(type_hint, ctx.value[member_name], member_name, ctx.field)
      elif style == A.union.Style.flat:
        child_context = ctx.push(type_hint, dict(ctx.value), None, ctx.field)
        t.cast(t.Dict, child_context.value).pop(discriminator_key)
      else:
        raise RuntimeError(f'bad style: {style!r}')
    else:
      child_context = ctx.push(type_hint, ctx.value, None, ctx.field)

    result = child_context.convert()

    if is_deserialize:
      return result
    else:
      if style == A.union.Style.nested:
        result = {discriminator_key: member_name, member_name: result}
      else:
        if not isinstance(result, t.MutableMapping):
          raise RuntimeError(f'union.Style.flat is not supported for non-object member types')
        result[discriminator_key] = member_name

    return result




import typing as t
from databind.core import Context, Converter, MapType


class MapConverter(Converter):

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, MapType)

    # # Catch subclasses of the typing.Dict generics. The instantiated generic with the type
    # # parameter will be stored in __orig_bases__.
    # dict_base = find_orig_base(context.type, Dict)
    # if dict_base:
    #   key_type, value_type = dict_base.__args__[0], dict_base.__args__[1]
    #   constructor = context.type
    #
    # # Otherwise, catch instances of the typing.List generic.
    # elif getattr(context.type, '__origin__', None) in (dict, Dict):
    #  # For the List generic.
    #  key_type, value_type = context.type.__args__[0], context.type.__args__[1]
    #  constructor = list
    #
    # else:
    #   raise RuntimeError(f'unsure how to handle type {type_repr(context.type)}')

    if not isinstance(ctx.value, t.Mapping):
      raise ctx.type_error(expected=t.Mapping)

    result = dict()  # TODO (@NiklasRosenstein): Check MapType.impl_hint

    for key, value in ctx.value.items():
      if ctx.type.key_type is not None:
        key = ctx.push(ctx.type.key_type, key, f'Key({key!r})', ctx.field).convert()
      if ctx.type.value_type is not None:
        value = ctx.push(ctx.type.value_type, value, key, ctx.field).convert()
      result[key] = value

    return result




import enum
import typing as t

from databind.core import BaseType, ConcreteType, Context, ConverterNotFound, Converter, ConverterProvider,  Direction
from databind.core import annotations as A


class EnumConverter(Converter, ConverterProvider):
  """
  Converter for enum values.

  * #enum.IntEnum subclasses are serialized to integers.
  * #enum.Enum subclasses are serialized to strings (from the enum value name).
  """

  def get_converter(self, type_: BaseType, direction: Direction) -> Converter:
    if isinstance(type_, ConcreteType) and issubclass(type_.type, enum.Enum):
      return self
    raise ConverterNotFound(type_, direction)

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, ConcreteType)
    assert issubclass(ctx.type.type, enum.Enum)

    if ctx.direction == Direction.serialize:
      if not isinstance(ctx.value, ctx.type.type):
        raise ctx.type_error(expected=ctx.type.type)
      if issubclass(ctx.type.type, enum.IntEnum):
        return ctx.value.value
      if issubclass(ctx.type.type, enum.Enum):
        name = ctx.value.name
        alias = ctx.annotations.get_field_annotation(ctx.type.type, name, A.alias)
        if alias and alias.aliases:
          return alias.aliases[0]
        return name

    elif ctx.direction == Direction.deserialize:
      if issubclass(ctx.type.type, enum.IntEnum):
        if not isinstance(ctx.value, int):
          raise ctx.type_error(expected=int)
        return ctx.type.type(ctx.value)
      if issubclass(ctx.type.type, enum.Enum):
        if not isinstance(ctx.value, str):
          raise ctx.type_error(expected=str)
        for enum_value in ctx.type.type:
          alias = ctx.annotations.get_field_annotation(ctx.type.type, enum_value.name, A.alias)
          if alias and ctx.value in alias.aliases:
            return enum_value
        try:
          return ctx.type.type[ctx.value]
        except KeyError:
          raise ctx.error(f'{ctx.value!r} is not a member of enumeration {ctx.type}')

    assert False




__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '1.2.0'

import datetime
import decimal
import io
import json
import pathlib
import typing as t
import uuid

from nr.parsing.date import duration

from databind.core import ObjectMapper, SimpleModule, ImplicitUnionType, ListType, MapType, ObjectType, OptionalType, SetType, UnionType
from .modules.any import AnyConverter
from .modules.collection import CollectionConverter
from .modules.datetime import DatetimeJsonConverter, DurationConverter
from .modules.decimal import DecimalJsonConverter
from .modules.enum import EnumConverter
from .modules.implicitunion import ImplicitUnionConverter
from .modules.map import MapConverter
from .modules.object import ObjectTypeConverter
from .modules.optional import OptionalConverter
from .modules.plain import PlainJsonConverter
from .modules.string import StringConverter
from .modules.union import UnionConverter

__all__ = [
  'JsonModule',
  'JsonType',
  'mapper',
  'load',
  'loads',
  'dump',
  'dumps',
]

T = t.TypeVar('T')
JsonType = t.Union[t.Mapping, t.Collection, str, int, float, bool, None]


class JsonModule(SimpleModule):
  """
  A composite of all modules for JSON de/serialization.
  """

  def __init__(self, name: str = None) -> None:
    super().__init__(name)

    self.add_converter_for_type(object, AnyConverter())
    self.add_converter_for_type(ObjectType, ObjectTypeConverter())
    self.add_converter_for_type(UnionType, UnionConverter())
    self.add_converter_for_type(bool, PlainJsonConverter())
    self.add_converter_for_type(float, PlainJsonConverter())
    self.add_converter_for_type(int, PlainJsonConverter())
    self.add_converter_for_type(bytes, PlainJsonConverter())
    self.add_converter_for_type(str, PlainJsonConverter())
    self.add_converter_for_type(uuid.UUID, StringConverter(lambda t, v: uuid.UUID(v)))
    self.add_converter_for_type(decimal.Decimal, DecimalJsonConverter())
    self.add_converter_for_type(datetime.date, DatetimeJsonConverter())
    self.add_converter_for_type(datetime.time, DatetimeJsonConverter())
    self.add_converter_for_type(datetime.datetime, DatetimeJsonConverter())
    self.add_converter_for_type(duration, DurationConverter())
    self.add_converter_for_type(OptionalType, OptionalConverter())
    self.add_converter_for_type(MapType, MapConverter())
    self.add_converter_for_type(ListType, CollectionConverter())
    self.add_converter_for_type(SetType, CollectionConverter())
    self.add_converter_for_type(ImplicitUnionType, ImplicitUnionConverter())
    self.add_converter_provider(EnumConverter())
    self.add_converter_provider(StringConverter(lambda t, v: t(v), matches=lambda t: issubclass(t, pathlib.PurePath)))


def mapper() -> ObjectMapper:
  return ObjectMapper(JsonModule(), name=__name__)


new_mapper = mapper  # backwards compatibility <=1.0.1


def load(
  data: t.Union[JsonType, t.TextIO],
  type_: t.Type[T],
  mapper: ObjectMapper = None,
  filename: str = None,
  annotations: t.List[t.Any] = None,
  options: t.List[t.Any] = None,
) -> T:

  if hasattr(data, 'read'):
    if not filename:
      filename = getattr(data, 'name', None)
    data = json.load(t.cast(t.TextIO, data))
  return (mapper or new_mapper()).deserialize(data, type_, filename=filename, annotations=annotations, settings=options)


def loads(
  data: str,
  type_: t.Type[T],
  mapper: ObjectMapper = None,
  filename: str = None,
  annotations: t.List[t.Any] = None,
  options: t.List[t.Any] = None,
) -> T:
  return load(io.StringIO(data), type_, mapper, filename, annotations, options)


def dump(
  value: T,
  type_: t.Type[T] = None,
  mapper: ObjectMapper = None,
  annotations: t.List[t.Any] = None,
  options: t.List[t.Any] = None,
  out: t.TextIO = None,
) -> JsonType:

  data = (mapper or new_mapper()).serialize(value, type_ or type(value), annotations=annotations, settings=options)
  if out is not None:
    json.dump(data, out)
  return data


def dumps(
  value: T,
  type_: t.Type[T] = None,
  mapper: ObjectMapper = None,
  annotations: t.List[t.Any] = None,
  options: t.List[t.Any] = None,
) -> str:
  return json.dumps(dump(value, type_, mapper, annotations, options))




import logging
import traceback
import typing as t
from databind.core import Context, ConversionError, Converter, ImplicitUnionType

logger = logging.getLogger(__name__)


def _get_log_level(logger: t.Optional[logging.Logger]) -> int:
  while logger and logger.level == 0:
    logger = logger.parent
  if logger:
    return logger.level
  return 0


def _indent_exc(exc: str) -> str:
  lines = []
  for index, line in enumerate(str(exc).splitlines()):
    if index > 0:
      line = '| ' + line
    lines.append(line)
  return '\n'.join(lines)


class ImplicitUnionConverter(Converter):

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, ImplicitUnionType)

    errors: t.List[str] = []
    tracebacks: t.List[str] = []
    is_debug = _get_log_level(logger) >= logging.DEBUG

    for type_ in ctx.type.types:
      try:
        return ctx.push(type_, ctx.value, None, ctx.field).convert()
      except ConversionError as exc:
        errors.append(f'{type_}: {exc}')
        if is_debug:
          tracebacks.append(traceback.format_exc())
    if is_debug:
      logger.debug(
        f'Error converting `{ctx.type}` ({ctx.direction.name}). This message is logged in '
        f'conjunction with a ConversionTypeError to provide information about the tracebacks '
        f'that have been caught when converting the individiual union members. This might not '
        f'indicate an error in the program if the exception is handled.\n'
        + '\n'.join(tracebacks))
    errors_text = '\n'.join(errors)
    raise ctx.error(_indent_exc(
      f'expected {ctx.type}, got `{type(ctx.value).__name__}`\n{errors_text}'))




import typing as t
from databind.core import Context, Converter, OptionalType


class OptionalConverter(Converter):

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, OptionalType)
    if ctx.value is None:
      return None
    return ctx.push(ctx.type.type, ctx.value, None, ctx.field).convert()



# This file was auto-generated by Shut. DO NOT EDIT
# For more information about Shut, check out https://pypi.org/project/shut/

from __future__ import print_function
import io
import os
import setuptools
import sys

command = sys.argv[1] if len(sys.argv) >= 2 else None

def _tempcopy(src, dst):
  import atexit, shutil
  if not os.path.isfile(dst):
    if not os.path.isfile(src):
      msg = '"{}" does not exist, and cannot copy it from "{}" either'.format(dst, src)
      # NOTE: In dist/build commands that are not invoked by Pip, we enforce that the license file
      #       must be present. See https://github.com/NiklasRosenstein/shut/issues/22
      if command and 'PIP_REQ_TRACKER' not in os.environ and ('build' in command or 'dist' in command):
        raise RuntimeError(msg)
      print('warning:', msg, file=sys.stderr)
      return
    shutil.copyfile(src, dst)
    atexit.register(lambda: os.remove(dst))


_tempcopy('../LICENSE.txt', 'LICENSE.txt')

readme_file = 'README.md'
if os.path.isfile(readme_file):
  with io.open(readme_file, encoding='utf8') as fp:
    long_description = fp.read()
else:
  print("warning: file \"{}\" does not exist.".format(readme_file), file=sys.stderr)
  long_description = None

requirements = [
  'databind.core >=1.2.0,<2.0.0',
  'typing_extensions >=3.10.0,<4.0.0',
  'nr.parsing.date >=1.0.1,<2.0.0',
  'nr.preconditions >=0.0.4,<1.0.0',
]

setuptools.setup(
  name = 'databind.json',
  version = '1.2.0',
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'De-/serialize Python dataclasses to or from JSON payloads. Compatible with Python 3.7 and newer.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein/python-databind',
  license = 'MIT',
  packages = setuptools.find_packages('src', ['test', 'test.*', 'tests', 'tests.*', 'docs', 'docs.*']),
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = {},
  tests_require = [],
  python_requires = '>=3.7.0,<4.0.0',
  data_files = [],
  entry_points = {},
  cmdclass = {},
  keywords = [],
  classifiers = [],
  zip_safe = False,
)




import dataclasses
import typing as t

from databind.core import BaseType, ConcreteType, Context, Converter, ConverterNotFound, ConverterProvider, Direction


@dataclasses.dataclass
class StringConverter(Converter, ConverterProvider):
  """
  This helper class makes it easy to define a converter for string based types in serialized form.
  """

  from_string: t.Callable[[t.Type, str], t.Any]
  to_string: t.Optional[t.Callable[[t.Type, t.Any], str]] = None
  matches: t.Optional[t.Callable[[t.Type], bool]] = None

  def get_converter(self, type_: BaseType, direction: Direction) -> Converter:
    if self.matches is None:
      raise RuntimeError('StringConverter cannot be as a ConverterProvider unless StringConverter.matches is set')
    if isinstance(type_, ConcreteType) and self.matches(type_.type):
      return self
    raise ConverterNotFound(type_, direction)

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, ConcreteType)

    if ctx.direction.is_deserialize():
      if not isinstance(ctx.value, str):
        raise ctx.type_error(expected=str)
      try:
        return self.from_string(ctx.type.type, ctx.value)
      except (TypeError, ValueError) as exc:
        raise ctx.error(str(exc))

    else:
      if not isinstance(ctx.value, ctx.type.type):
        raise ctx.type_error(expected=ctx.type.type)
      if self.to_string is None:
        return str(ctx.value)
      else:
        return self.to_string(ctx.type.type, ctx.value)




import typing as t
from databind.core import Context, Direction, Converter, CollectionType


class CollectionConverter(Converter):

  def __init__(self, json_type: t.Type[t.Collection] = list) -> None:
    self.json_type = json_type

  def convert(self, ctx: Context) -> t.Any:
    assert isinstance(ctx.type, CollectionType)

    if ctx.direction == Direction.serialize:
      if not isinstance(ctx.value, ctx.type.python_type):
        raise ctx.type_error(expected=ctx.type.python_type)
      python_type = self.json_type

    elif ctx.direction == Direction.deserialize:
      if not isinstance(ctx.value, t.Collection) or isinstance(ctx.value, (str, bytes, bytearray, memoryview)):
        raise ctx.type_error(expected=t.Collection)
      python_type = ctx.type.python_type

    else:
      assert False, ctx.direction

    return python_type(  # type: ignore
      ctx.push(ctx.type.item_type, val, idx, ctx.field).convert()
      for idx, val in enumerate(ctx.value))




"""
Provides the #DecimalModule for decimal value de/serialization.
"""

import decimal
import typing as t
from databind.core import annotations as A, Context, ConcreteType, Direction, Converter
from nr import preconditions
from nr.optional import Optional


class DecimalJsonConverter(Converter):

  def convert(self, ctx: Context) -> t.Any:
    preconditions.check_instance_of(ctx.type, ConcreteType)
    preconditions.check_argument(t.cast(ConcreteType, ctx.type).type is decimal.Decimal, 'must be Decimal')
    context = Optional(ctx.get_annotation(A.precision))\
      .map(lambda b: b.to_context()).or_else(None)
    fieldinfo = ctx.get_annotation(A.fieldinfo) or A.fieldinfo()

    if ctx.direction == Direction.deserialize:
      if (not fieldinfo.strict and isinstance(ctx.value, (int, float))) or isinstance(ctx.value, str):
        return decimal.Decimal(ctx.value, context)
      raise ctx.type_error(expected='str')

    else:
      if not isinstance(ctx.value, decimal.Decimal):
        raise ctx.type_error(expected=decimal.Decimal)
      return str(ctx.value)
